import dataclasses
import io
import glob
import pathlib
import logging
import sys
import subprocess
import shutil
import abc

from .os import shutil_which

from typing import (
    Optional,
    Literal,
)

logger = logging.getLogger(__name__)

@dataclasses.dataclass
class Project:
    source_dir : pathlib.Path
    build_dir : pathlib.Path
    dest_dir : pathlib.Path

@dataclasses.dataclass
class Dependency:
    name: str
    mode : Literal['pyproject', 'meson', 'meson-python', 'm']
    source_path : pathlib.Path
    args: Optional[list[str]] = None

@dataclasses.dataclass
class DistSettings:
    wheel_dir : pathlib.Path
    python_path: pathlib.Path
    env_path: pathlib.Path

class CLI(abc.ABC):
    @property
    @abc.abstractmethod
    def dist_settings(self) -> DistSettings:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def projects(self) -> dict[str, Project]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dependencies(self) -> dict[str, Dependency]:
        raise NotImplementedError

    def deploy_fetch_dist(
        self,
        force: bool,
    ) -> None:
        for k, o in self.dependencies.items():
            whl_glob = self.dist_settings.wheel_dir / ('*%s*.whl' % o.name.replace('.', '_'))
            if len(glob.glob(
                str(whl_glob)
            )) == 0 or force:
                if o.source_path.exists():
                    def whl_files_get() -> list[dict[str, Any]]:
                        return [
                            dict(
                                path=o,
                                stat=os.stat(o).st_mtime,
                            )
                            for o in glob.glob(
                                str(whl_glob)
                            )
                        ]

                    present_files = whl_files_get()

                    if o.mode == 'm':
                        if (o.source_path / 'm.py').exists():
                            cmd = [
                                sys.executable,
                                o.source_path / 'm.py',
                                'deploy:wheel',
                                '-o', self.settings.wheel_dir,
                            ]

                            if not o.args is None:
                                cmd.extend(o.args)

                            subprocess.check_call(cmd)
                    else:
                        raise NotImplementedError

                    updated_files = whl_files_get()

                    def index_get(o: dict[str, Any]) -> tuple[Any, ...]:
                        return (o['path'], o['stat'])

                    present_files_index = {
                        index_get(o) : o
                        for o in present_files
                    }

                    new_files : list[dict[str, Any]] = []

                    for o in updated_files:
                        entry_index = index_get(o)

                        if not entry_index in present_files_index:
                            new_files.append(o)

                    if len(new_files) == 0:
                        raise NotImplementedError

                    latest_file = sorted(
                        new_files,
                        key=lambda x: x['stat']
                    )[-1]

                    subprocess.check_call([
                        self.dist_settings.python_path,
                        '-m', 'pip',
                        'install', 
                        latest_file['path'],
                    ])

    def deploy_wheel(
        self,
        project_name: str,
        argv: Optional[list[str]] = None,
        output_dir: Optional[pathlib.Path] = None,
        force: Optional[bool] = None,
    ) -> None:
        project = self.projects[project_name]

        # subprocess.check_call([
        #     sys.argv[0],
        #     # sys.executable,
        #     '-p', options.project,
        #     Command.meson_setup.value,
        # ])

        assert argv is None or len(argv) == 0

        self.meson_install(
            project_name=project_name,
            force=force,
        )

        cmd = [
            sys.executable,
            '-m',
            'build',
            '-w', '-n',
            '-Csetup-args=-Dmodes=pyproject',
            '-Csetup-args=-Dinstall_path=%s' % str(project.dest_dir),
            # '-Cbuild-dir=%s' % str(project.build_dir),
            str(project.source_dir),
        ]

        if not output_dir is None:
            cmd.extend(['-o', str(output_dir)])

        subprocess.check_call(cmd)

    def meson_install(
        self,
        project_name: str,
        force: Optional[bool] = None,
        argv: Optional[list[str]] = None,
    ) -> None:
        project = self.projects[project_name]

        if force is None:
            force = False

        if argv is None:
            argv = []

        if force and project.dest_dir.exists():
            shutil.rmtree(project.dest_dir)

        subprocess.check_call([
            shutil_which('meson', True,),
            'install',
            '-C',
            project.build_dir,
            '--destdir', project.dest_dir,
            *argv,
        ])

        for o in glob.glob(
            str(project.dest_dir / 'lib' / 'pkgconfig' / '*.pc'),
            recursive=True,
        ):
            logger.info(dict(
                path=o,
                action='patch prefix',
            ))

            with io.open(o, 'r') as f:
                content = f.read()

            with io.open(o, 'w') as f:
                f.write(
                    content.replace('prefix=/', 'prefix=${pcfiledir}/../../')
                )
    def ninja(
        self,
        project_name: str,
        argv: Optional[list[str]] = None,
    ) -> None:
        project = self.projects[project_name]

        if argv is None:
            argv = []

        subprocess.check_call([
            shutil_which('ninja', True),
            '-C',
            str(project.build_dir),
            *argv,
        ])

    def meson_compile(
        self,
        project_name: str,
        argv: Optional[list[str]] = None,
    ) -> None:
        project = self.projects[project_name]

        if argv is None:
            argv = []

        subprocess.check_call([
            shutil_which('meson', True,),
            'compile',
            '-C',
            project.build_dir,
            *argv,
        ])

    def meson_setup(
        self,
        project_name: str,
        force: bool,
        argv: Optional[list[str]] = None,
    ) -> None:
        project = self.projects[project_name]

        if argv is None:
            argv = []

        if force:
            if project.build_dir.exists():
                logger.info(dict(action='removing build dir', path=project.build_dir))
                shutil.rmtree(project.build_dir)

        cmd = [
            shutil_which('meson', True,),
            'setup',
            str(project.source_dir),
            str(project.build_dir),
            '-Dmodes=["meson"]',
            # '-Dpkgconfig.relocatable=true',
            '-Dprefix=/',
            *argv,
        ]

        logger.info(dict(cmd=cmd))

        subprocess.check_call(cmd)